module mytype
  implicit none
  integer, parameter:: dp=kind(1.0D0)
  integer, parameter:: qp=2*dp
end module mytype

module matrix
  use mytype
  implicit none

  interface operator (.mm.)
     module procedure mm, vm, mv, vv
  end interface

  interface operator (.tmm.)
     module procedure tmm
  end interface

  interface operator (.mmt.)
     module procedure mmt
  end interface

  interface operator (.cb.)
     module procedure cbind
  end interface

  interface operator (.rb.)
     module procedure rbind
  end interface

  interface operator (.kp.)
     module procedure kron_mm, kron_mv, kron_vm
  end interface

  interface operator (.mmd.)
     module procedure mmdiag_l, mmdiag_r
  end interface

  interface operator (.tmmd.)
     module procedure mmdiagt_l
  end interface

  interface operator (.mmdt.)
     module procedure mmdiagt_r
  end interface

  interface printmat
     module procedure printmat_i, printmat_d, printvec_i, printvec_d
  end interface 

  interface diag
     module procedure ddiag, diagonal
  end interface

  interface qsort
     module procedure iqsort, dqsort
  end interface

  interface poisfit
     module procedure poisfito, poisfitn
  end interface

  interface symsolve
     module procedure symsolve_v, symsolve_m
  end interface

  interface fsymsolve
     module procedure fsymsolve_v, fsymsolve_m
  end interface

  interface rsymsolve
     module procedure rsymsolve_v, rsymsolve_m
  end interface

  interface sqrsolve
     module procedure sqrsolve_v, sqrsolve_m
  end interface

  interface fsqrsolve
     module procedure fsqrsolve_v, fsqrsolve_m
  end interface

  interface trisolve
     module procedure trisolve_v, trisolve_m
  end interface

contains      

  subroutine readmat(A,data)
    implicit none
    real(dp), dimension(:,:), intent(inout):: A
    character(len=*), intent(in):: data
    integer:: i
    open(1,file=data,status='old',access='sequential',action='read')
    do i=1,size(A,1)
       read(1,*) A(i,:)
    end do 
    close(1)
  end subroutine readmat

  subroutine printmat_d(A)
    implicit none
    real(dp), dimension(:,:):: A
    integer:: i, j
    do i=1,size(A,1)
       write(*,fmt="(a2)",advance="no") "|"
       do j=1,size(A,2)
          write(*,fmt="(f9.4)",advance="no") A(i,j)
       end do
       write(*,fmt="(a2)") " |"
    end do
    write(*,*)
  end subroutine printmat_d 

  subroutine printmat_i(A)
    implicit none
    integer, dimension(:,:):: A
    integer:: i, j
    do i=1,size(A,1)
       write(*,fmt="(a2)",advance="no") "|"
       do j=1,size(A,2)
          write(*,fmt="(i5)",advance="no") A(i,j)
       end do
       write(*,fmt="(a2)") " |"
    end do
    write(*,*)
  end subroutine printmat_i 

  subroutine printvec_d(A)
    implicit none
    real(dp), dimension(:):: A
    integer:: i
    write(*,fmt="(a2)",advance="no") "|"
    do i=1,size(A)
       write(*,fmt="(f9.4)",advance="no") A(i)
    end do
    write(*,fmt="(a2)") " |"
    write(*,*)
  end subroutine printvec_d

  subroutine printvec_i(A)
    implicit none
    integer, dimension(:):: A
    integer:: i
    write(*,fmt="(a2)",advance="no") "|"
    do i=1,size(A)
       write(*,fmt="(i5)",advance="no") A(i)
    end do
    write(*,fmt="(a2)") " |"
    write(*,*)
  end subroutine printvec_i

  function mm(A,B) result(out)
    implicit none
    real(dp), dimension(:,:), intent(in):: A, B
    real(dp), dimension(size(A,1),size(B,2)):: out
    call dgemm_f95(A,B,out,'N','N')
  end function mm

  function tmm(A,B) result(out)
    implicit none
    real(dp), dimension(:,:), intent(in):: A, B
    real(dp), dimension(size(A,2),size(B,2)):: out
    call dgemm_f95(A,B,out,'T','N')
  end function tmm

  function mmt(A,B) result(out)
    implicit none
    real(dp), dimension(:,:), intent(in):: A, B
    real(dp), dimension(size(A,1),size(B,1)):: out
    call dgemm_f95(A,B,out,'N','T')
  end function mmt

  function mv(A,B) result(out)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(:), intent(in):: B
    real(dp), dimension(size(A,1)):: out
    call dgemv_f95(A,B,out,'N')
  end function mv

  function vm(A,B) result(out)
    implicit none
    real(dp), dimension(:,:), intent(in)::  B
    real(dp), dimension(:), intent(in):: A
    real(dp), dimension(size(B,2)):: out
    call dgemv_f95(B,A,out,'T')
  end function vm

  function vv(A,B) result(out)
    implicit none
    real(dp), dimension(:), intent(in)::  B
    real(dp), dimension(:), intent(in):: A
    real(dp):: out
    out=dot_product(A,B)
  end function vv

  SUBROUTINE dgemm_f95(a,b,c,t1,t2)
    ! .. Use Statements ..
    ! .. Parameters ..
    REAL (dp), PARAMETER :: zero = 0.0_dp
    REAL (dp), PARAMETER :: one = 1.0_dp
    ! .. Array Arguments ..
    REAL (dp), INTENT (IN) :: a(:,:), b(:,:)
    CHARACTER, INTENT (IN) :: t1, t2
    REAL (dp), INTENT (OUT) :: c(:,:)
    ! .. Local Scalar ..
    INTEGER :: k, m, n
    ! .. External Procedures ..
    EXTERNAL dgemm  
    m = SIZE(c,1)
    n = SIZE(c,2)
    IF (t1 == 'T'.and. t2 == 'T') then
       k = SIZE(a,1)
       CALL dgemm(t1,t2,m,n,k,one,a,k,b,n,zero,c,m)
    ELSEIF (t1 == 'T'.and. t2 == 'N') then
       k = SIZE(a,1)
       CALL dgemm(t1,t2,m,n,k,one,a,k,b,k,zero,c,m)
    ELSEIF (t1 == 'N'.and. t2 == 'T') then
       k = SIZE(a,2)
       CALL dgemm(t1,t2,m,n,k,one,a,m,b,n,zero,c,m)
    ELSE
       k = SIZE(a,2)
       CALL dgemm(t1,t2,m,n,k,one,a,m,b,k,zero,c,m)
    END IF
  END SUBROUTINE dgemm_f95
  
  SUBROUTINE dgemv_f95(a,b,c,t1)
    ! .. Use Statements ..
    ! .. Parameters ..
    REAL (dp), PARAMETER :: one = 1.0_dp
    REAL (dp), PARAMETER :: zero = 0.0_dp
    ! .. Array Arguments ..
    CHARACTER, INTENT (IN) :: t1
    REAL (dp), INTENT (IN) :: a(:,:), b(:)
    REAL (dp), INTENT (OUT) :: c(:)
    ! .. Local Scalar ..
    INTEGER :: m, n
    ! .. External Procedures ..
    EXTERNAL dgemv  
    m = SIZE(a,1)
    n = SIZE(a,2)
    CALL dgemv(t1,m,n,one,a,m,b,1,zero,c,1)
  END SUBROUTINE dgemv_f95

  function mmdiag_l(A,diag)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2)), intent(in):: diag
    real(dp), dimension(size(A,1),size(A,2)):: mmdiag_l
    integer:: i
    do i=1,size(A,2)
       mmdiag_l(:,i)=A(:,i)*diag(i)
    end do
  end function mmdiag_l

  function mmdiag_r(diag,A)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,1)), intent(in):: diag
    real(dp), dimension(size(A,1),size(A,2)):: mmdiag_r
    integer:: i
    do i=1,size(A,1)
       mmdiag_r(i,:)=A(i,:)*diag(i)
    end do
  end function mmdiag_r

  function mmdiagt_l(A,diag)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,1)), intent(in):: diag
    real(dp), dimension(size(A,2),size(A,1)):: mmdiagt_l
    integer:: i
    do i=1,size(A,1)
       mmdiagt_l(:,i)=A(i,:)*diag(i)
    end do
  end function mmdiagt_l

  function mmdiagt_r(diag,A)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2)), intent(in):: diag
    real(dp), dimension(size(A,2),size(A,1)):: mmdiagt_r
    integer:: i
    do i=1,size(A,2)
       mmdiagt_r(i,:)=A(:,i)*diag(i)
    end do
  end function mmdiagt_r

  function ddiag(A)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    integer:: i
    real(dp), dimension(size(A,1)):: ddiag
    do i=1,size(A,1)
       ddiag(i)=A(i,i)
    end do
  end function ddiag

  function diagonal(a)
    implicit none
    real(dp), dimension(:), intent(in):: a
    integer:: i
    real(dp), dimension(size(a),size(a)):: diagonal
    diagonal=0.0_dp
    do i=1,size(a)
       diagonal(i,i)=a(i)
    end do
  end function diagonal

  subroutine det(A,n,dd)
  !  
  ! Compute the determinant of a square symmetric positive definite
  ! matrix.
  !
  ! Variables::
  !
  !       in: A - nxn matrix 
  !       in: n - dimensions of A
  !      out: dd - the determinant
  ! 
    implicit none
    integer, intent(in):: n
    real(dp), intent(in):: A(n,n)
    real(dp), intent(out):: dd
    real(dp):: B(n,n)
    integer:: info
    B=A
    call dpotrf('U',n,B,n,info)
    if(info/=0) then
       write(*,*) 'det: failure in routine'
       dd=0.0_dp
       return
    end if
    dd=(product(ddiag(B)))**2
  end subroutine det

  subroutine ldet(A,n,dd)
  !  
  ! Compute the log of the determinant of a square symmetric positive 
  ! definite  matrix.
  !
  ! Variables::
  !
  !       in: A - nxn matrix 
  !       in: n - dimensions of A
  !      out: dd - the log determinant
  ! 
    implicit none
    integer, intent(in):: n
    real(dp), intent(in):: A(n,n)
    real(dp), intent(out):: dd
    real(dp):: B(n,n)
    integer:: info
    B=A
    call dpotrf('U',n,B,n,info)
    if(info/=0) then
       write(*,*) 'det: failure in routine'
       dd=0.0_dp
       return
    end if
    dd=2*sum(log(ddiag(B)))
  end subroutine ldet

  subroutine chol(A,B,n,UPLO)
  !
  ! Compute the Choleski decompostion of A: 
  !         
  ! Variables:
  !           in:   n - dimension of A and B
  !           in:   A -  nxn symmetric matrix
  !           out:  B -  nxn lower trianglar matrix
  !                      i.e. A=t(R).mm.R  or A=L.mm.t(LL)
  !           in: UPLO- optional 'U' upper tri, 'L' lower tri.  If not
  !                     present then 'U'
  !
    implicit none
    integer, intent(in):: n
    real(dp), intent(in):: A(n,n)
    real(dp), intent(out):: B(n,n)
    character, optional, intent(in):: UPLO
    character:: UPL
    integer:: info
    if (.not.present(UPLO)) then
       UPL='U'
    else
       if(UPLO/='U' .and. UPLO/='L') then
          stop 'chol: Must be one of U or L!'
       end if
       UPL=UPLO
    end if
    B=A
    call dpotrf(UPL,n,B,n,info)
    if(info/=0) then
       stop 'chol: failure in routine'
    end if
  end subroutine chol

  function eye(n)
    implicit none
    integer, intent(in):: n
    integer:: i
    real(dp), dimension(n,n):: eye
    real(dp), dimension(n):: a
    a=(/ (1.0_dp,i=1,n) /)
    eye=diagonal(a)
  end function eye

  function rbind(A,B)
    implicit none
    real(dp), dimension(:,:), intent(in):: A, B
    real(dp), dimension((size(A,1)+size(B,1)),size(A,2)):: rbind
    rbind(1:size(A,1),1:size(A,2))=A
    rbind((size(A,1)+1):(size(A,1)+size(B,1)),1:size(A,2))=B
  end function rbind

  function cbind(A,B)
    implicit none
    real(dp), dimension(:,:), intent(in):: A, B
    real(dp), dimension(size(A,1),(size(A,2)+size(B,2))):: cbind
    cbind(1:size(A,1),1:size(A,2))=A
    cbind(1:size(A,1),(size(A,2)+1):(size(A,2)+size(B,2)))=B
  end function cbind    

  function vech(X)
    use mytype
    implicit none
    real(dp), intent(in):: X(:,:)
    real(dp):: vech((size(X,1)*(size(X,1)+1))/2)
    integer:: i, nn, ll, uu
    nn=size(X,1)
    if (size(X,2)/=nn) stop 'vech: Matrix X must be symmetric'
    uu=0
    do i=1,nn
       ll=uu+1
       uu=uu+nn-i+1
       vech(ll:uu)=X(i:nn,i)
    end do
  end function vech

  function dup_mat(n) result(DD)
    use mytype
    implicit none
    integer, intent(in):: n
    real(dp):: DD(n*n,n*(n+1)/2)
    integer:: count, i, j
    DD=0.0_dp
    count=0
    count = 0;
    do j = 1,n
       dd((j - 1) * n + j, count + j) = 1.0_dp
       do i = (j + 1),n
          dd((j - 1) * n + i, count + i) = 1;
          dd((i - 1) * n + j, count + i) = 1;
       end do
       count = count + n - j;
    end do
  end function dup_mat

  subroutine kron_prod(a,ia,ma,na,b,ib,mb,nb,pk,ik)
    use mytype
    implicit none
    real(dp) a(*),b(*),pk(*)
    integer ia,ma,na,ib,mb,nb,ik,ka,kb,kk,ka1,kk1,ja,jb,i
    ka1=1-ia
    kk1=-nb
    do ja=1,na
       kb=1
       ka1=ka1+ia
       kk1=kk1+nb
       do jb=1,nb
          ka=ka1
          kk=1+(jb-1+kk1)*ik
          do i=1,ma
             call dcopy(mb,b(kb),1,pk(kk),1)
             call dscal(mb,a(ka),pk(kk),1)
             kk=kk+mb
             ka=ka+1
          end do
          kb=kb+ib
       end do
    end do
    return
  end subroutine kron_prod

  function kron_mm(A, B) result(out)
    implicit none
    real(dp), dimension(:,:), intent(in):: A, B
    real(dp), dimension(size(A,1)*size(B,1),size(A,2)*size(B,2)):: out
    integer, dimension(2):: an, bn
    an=shape(A)
    bn=shape(B)
    call kron_prod(A,an(1),an(1),an(2),B,bn(1),bn(1),bn(2),out,an(1)*bn(1))
  end function kron_mm

  function kron_vm(A, B) result(out)
    implicit none
    real(dp), dimension(:), intent(in):: A
    real(dp), dimension(:,:), intent(in):: B
    real(dp), dimension(size(A)*size(B,1),size(B,2)):: out
    integer:: an, bn(2)
    an=size(A)
    bn=shape(B)
    call kron_prod(A,an,an,1,B,bn(1),bn(1),bn(2),out,an*bn(1))
  end function kron_vm

  function kron_mv(A, B) result(out)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(:), intent(in):: B
    real(dp), dimension(size(A,1)*size(B,1),size(A,2)):: out
    integer:: an(2), bn
    an=shape(A)
    bn=size(B)
    call kron_prod(A,an(1),an(1),an(2),B,bn,bn,1,out,an(1)*bn)
  end function kron_mv

  function trace(A)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp):: trace
    trace=sum(ddiag(A))
  end function trace

  SUBROUTINE iqsort(a)
    !     NON-RECURSIVE STACK VERSION OF QUICKSORT FROM N.WIRTH'S PASCAL
    !     BOOK, 'ALGORITHMS + DATA STRUCTURES = PROGRAMS'.
    !     SINGLE PRECISION, ALSO CHANGES THE ORDER OF THE ASSOCIATED ARRAY T.
    IMPLICIT NONE
    INTEGER, DIMENSION(:), INTENT(INOUT)    :: a
    INTEGER:: n
    !     Local Variables
    INTEGER                :: i, j, k, l, r, s, stackl(15), stackr(15), w, x
    
    n=size(a)
    s = 1
    stackl(1) = 1
    stackr(1) = n

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.
    
10  CONTINUE
    l = stackl(s)
    r = stackr(s)
    s = s - 1

    !     KEEP SPLITTING A(L), ... , A(R) UNTIL L >= R.

20  CONTINUE
    i = l
    j = r
    k = (l+r) / 2
    x = a(k)
    
    !     REPEAT UNTIL I > J.

    DO
       DO
          IF (a(i).LT.x) THEN                ! Search from lower end
             i = i + 1
             CYCLE
          ELSE
             EXIT
          END IF
       END DO
       
       DO
          IF (x.LT.a(j)) THEN                ! Search from upper end
             j = j - 1
             CYCLE
          ELSE
             EXIT
          END IF
       END DO
       
       IF (i.LE.j) THEN                     ! Swap positions i & j
          w = a(i)
          a(i) = a(j)
          a(j) = w
          i = i + 1
          j = j - 1
          IF (i.GT.j) EXIT
       ELSE
          EXIT
       END IF
    END DO
    
    IF (j-l.GE.r-i) THEN
       IF (l.LT.j) THEN
          s = s + 1
          stackl(s) = l
          stackr(s) = j
       END IF
       l = i
    ELSE
       IF (i.LT.r) THEN
          s = s + 1
          stackl(s) = i
          stackr(s) = r
       END IF
       r = j
    END IF

    IF (l.LT.r) GO TO 20
    IF (s.NE.0) GO TO 10

    RETURN
  END SUBROUTINE iqsort

  SUBROUTINE dqsort(a)

    !     NON-RECURSIVE STACK VERSION OF QUICKSORT FROM N.WIRTH'S PASCAL
    !     BOOK, 'ALGORITHMS + DATA STRUCTURES = PROGRAMS'.
    !     SINGLE PRECISION, ALSO CHANGES THE ORDER OF THE ASSOCIATED ARRAY T.

    IMPLICIT NONE

    REAL(dp), DIMENSION(:), INTENT(INOUT)    :: a
    INTEGER:: n

    !     Local Variables

    INTEGER                :: i, j, k, l, r, s, stackl(15), stackr(15)
    REAL(dp)               :: w, x

    n=size(a)
    s = 1
    stackl(1) = 1
    stackr(1) = n

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.
    
10  CONTINUE
    l = stackl(s)
    r = stackr(s)
    s = s - 1

    !     KEEP SPLITTING A(L), ... , A(R) UNTIL L >= R.

20  CONTINUE
    i = l
    j = r
    k = (l+r) / 2
    x = a(k)
    
    !     REPEAT UNTIL I > J.

    DO
       DO
          IF (a(i).LT.x) THEN                ! Search from lower end
             i = i + 1
             CYCLE
          ELSE
             EXIT
          END IF
       END DO
       
       DO
          IF (x.LT.a(j)) THEN                ! Search from upper end
             j = j - 1
             CYCLE
          ELSE
             EXIT
          END IF
       END DO
       
       IF (i.LE.j) THEN                     ! Swap positions i & j
          w = a(i)
          a(i) = a(j)
          a(j) = w
          i = i + 1
          j = j - 1
          IF (i.GT.j) EXIT
       ELSE
          EXIT
       END IF
    END DO
    
    IF (j-l.GE.r-i) THEN
       IF (l.LT.j) THEN
          s = s + 1
          stackl(s) = l
          stackr(s) = j
       END IF
       l = i
    ELSE
       IF (i.LT.r) THEN
          s = s + 1
          stackl(s) = i
          stackr(s) = r
       END IF
       r = j
    END IF

    IF (l.LT.r) GO TO 20
    IF (s.NE.0) GO TO 10

    RETURN
  END SUBROUTINE dqsort

  function inverse(A,symm)
    implicit none
    logical, intent(in):: symm
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,1),size(A,2)):: inverse
    integer:: info
    inverse=eye(size(A,1))
    if(symm) then
       call fsymsolve_m(A,inverse,inverse,info)
    else
       call fsqrsolve_m(A,inverse,inverse,info)
    end if
    if (info/=0) then
       stop 'inverse: error in routine!'
    end if
  end function inverse

  function finverse(A,info) result(inverse)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    integer, intent(out):: info
    real(dp), dimension(size(A,1),size(A,2)):: inverse
    integer:: n
    n=size(A,1)
    inverse=eye(n)
    call fsymsolve_m(A,inverse,inverse,info)
  end function finverse

  function rinverse(A) result(inverse)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,1),size(A,2)):: inverse
    integer:: n
    n=size(A,1)
    inverse=eye(n)
    call rsymsolve_m(A,inverse,inverse)
  end function rinverse

  subroutine symsolve_v(A,b,syslv,ldet)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2)), intent(in):: b
    real(dp), dimension(size(A,2)), intent(out):: syslv
    real(dp), optional, intent(out):: ldet
    integer:: info
    if (present(ldet)) then
       call fsymsolve_v(A,b,syslv,info,ldet)
    else
       call fsymsolve_v(A,b,syslv,info)
    end if
    if (info/=0) then
       stop 'symsolve_v: error in routine!'
    end if
  end subroutine symsolve_v

  subroutine symsolve_m(A,B,syslv,ldet)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(:,:), intent(in):: B
    real(dp), dimension(size(A,2),size(B,2)), intent(out):: syslv
    real(dp), optional, intent(out):: ldet
    integer:: info
    if (present(ldet)) then
       call fsymsolve_m(A,B,syslv,info,ldet)
    else
       call fsymsolve_m(A,B,syslv,info)
    end if
    if (info/=0) then
       stop 'symsolve_m: error in routine!'
    end if
  end subroutine symsolve_m

  subroutine fsymsolve_v(A,b,syslv,info,ldet)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2)), intent(in):: b
    real(dp), dimension(size(A,2)), intent(out):: syslv
    integer, intent(out):: info
    real(dp), optional, intent(out):: ldet
    real(dp), dimension(size(A,1),size(A,2)):: D
    real(dp), dimension(size(A,2),1):: c
    integer:: n, nrhs, ipiv(size(A,1)), lwrk
    real(dp), allocatable:: wrk(:)
    real(dp):: w(1)
    n=size(A,1)
    nrhs=1
    if(n/=size(A,2)) then
       stop 'Matrix must be square'
    end if
    D=A
    c=reshape(b,(/n,1/))
    lwrk=-1
    call dsysv('U',n,nrhs,D,n,ipiv,c,n,w,lwrk,info)
    lwrk=w(1)
    allocate(wrk(lwrk))
    call dsysv('U',n,nrhs,D,n,ipiv,c,n,wrk,lwrk,info)
    if (present(ldet)) then
       ldet=sum(log(ddiag(D)))
    end if
    syslv=pack(c,.TRUE.)
    deallocate(wrk)
  end subroutine fsymsolve_v

  subroutine fsymsolve_m(A,B,syslv,info,ldet)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(:,:), intent(in):: B
    real(dp), dimension(size(A,2),size(B,2)), intent(out):: syslv
    integer, intent(out):: info
    real(dp), optional, intent(out):: ldet
    real(dp), dimension(size(A,1),size(A,2)):: D
    real(dp), dimension(size(A,2),size(B,2)):: C
    integer:: n, nrhs, ipiv(size(A,1)), lwrk
    real(dp), allocatable:: wrk(:)
    real(dp):: w(1)
    n=size(A,1)
    nrhs=size(B,2)
    if(n/=size(A,2)) then
       stop 'Matrix must be square'
    end if
    D=A
    C=B
    lwrk=-1
    call dsysv('U',n,nrhs,D,n, ipiv, C, n, w, lwrk, info)
    lwrk=w(1)
    allocate(wrk(lwrk))
    call dsysv('U',n,nrhs,D,n, ipiv, C, n, wrk, lwrk, info)
    if (present(ldet)) then
       ldet=sum(log(ddiag(D)))
    end if
    syslv=C
    deallocate(wrk)
  end subroutine fsymsolve_m

  subroutine sqrsolve_v(A,b,syslv,ldet)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2)), intent(in):: b
    real(dp), dimension(size(A,2)), intent(out):: syslv
    real(dp), optional, intent(out):: ldet
    integer:: info
    if (present(ldet)) then
       call fsqrsolve_v(A,b,syslv,info,ldet)
    else
       call fsqrsolve_v(A,b,syslv,info)
    end if
    if (info/=0) then
       stop 'sqrsolve_v: error in routine!'
    end if    
  end subroutine sqrsolve_v

  subroutine sqrsolve_m(A,B,syslv,ldet)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(:,:), intent(in):: B
    real(dp), dimension(size(A,2),size(B,2)), intent(out):: syslv
    real(dp), optional, intent(out):: ldet
    integer:: info
    if (present(ldet)) then
       call fsqrsolve_m(A,B,syslv,info,ldet)
    else
       call fsqrsolve_m(A,B,syslv,info)
    end if
    if (info/=0) then
       stop 'symsolve_m: error in routine!'
    end if
  end subroutine sqrsolve_m

  subroutine fsqrsolve_v(A,b,syslv,info,ldet)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2)), intent(in):: b
    real(dp), dimension(size(A,2)), intent(out):: syslv
    integer, intent(out):: info
    real(dp), optional, intent(out):: ldet
    real(dp), dimension(size(A,1),size(A,2)):: D
    real(dp), dimension(size(A,2),1):: c
    integer:: n, nrhs, ipiv(size(A,1))
    n=size(A,1)
    nrhs=1
    if(n/=size(A,2)) then
       stop 'Matrix must be square'
    end if
    D=A
    c=reshape(b,(/n,1/))
    call dgesv(n, nrhs, D, n, ipiv, c, n, info)
    if (present(ldet)) then
       ldet=sum(log(ddiag(D)))
    end if
    syslv=pack(c,.TRUE.)
  end subroutine fsqrsolve_v

  subroutine fsqrsolve_m(A,b,syslv,info,ldet)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(:,:), intent(in):: B
    real(dp), dimension(size(A,2),size(B,2)), intent(out):: syslv
    integer, intent(out):: info
    real(dp), optional, intent(out):: ldet
    real(dp), dimension(size(A,1),size(A,2)):: D
    real(dp), dimension(size(A,2),size(B,2)):: C
    integer:: n, nrhs, ipiv(size(A,1))
    n=size(A,1)
    nrhs=size(B,2)
    if(n/=size(A,2)) then
       stop 'Matrix must be square'
    end if
    D=A
    C=B
    call dgesv(n, nrhs, D, n, ipiv, C, n, info)
    if (present(ldet)) then
       ldet=sum(log(ddiag(D)))
    end if
    syslv=C
  end subroutine fsqrsolve_m

  subroutine rsymsolve_v(A,b,syslv)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2)), intent(in):: b
    real(dp), dimension(size(A,2)), intent(out):: syslv
    real(dp), dimension(size(A,1),size(A,2)):: D
    real(dp), dimension(size(A,2)):: c
    if(size(A,1)/=size(A,2)) then
       stop 'Matrix must be square'
    end if
    call symeigen(A,c,D)
    syslv=b.mm.D
    where (abs(c) < 1.0e-8_dp)
       c=0.0_dp
    elsewhere
       c=1/c
    end where
    syslv=c*syslv
    syslv=D.mm.syslv
  end subroutine rsymsolve_v

  subroutine rsymsolve_m(A,b,syslv)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(:,:), intent(in):: B
    real(dp), dimension(size(A,2),size(B,2)), intent(out):: syslv
    real(dp), dimension(size(A,1),size(A,2)):: D
    real(dp), dimension(size(A,2)):: c
    if(size(A,1)/=size(A,2)) then
       stop 'Matrix must be square'
    end if
    call symeigen(A,c,D)
    syslv=D.tmm.b
    where (abs(c) < 1.0e-8_dp)
       c=0.0_dp
    elsewhere
       c=1/c
    end where
    syslv=c.mmd.syslv
    syslv=D.mm.syslv
  end subroutine rsymsolve_m

  subroutine trisolve_v(A,b,trislv,trans)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2)), intent(in):: b
    character, intent(in):: trans
    real(dp), dimension(size(A,2)), intent(out):: trislv
    real(dp), dimension(size(A,1),size(A,2)):: D
    real(dp), dimension(size(A,2),1):: c
    integer:: n, info
    if(size(A,1)/=size(A,2)) then
       stop 'Matrix must be square'
    end if
    n=size(A,1)
    D=A
    c=reshape(b,(/size(A,1),1/))
    call dtrtrs('U',trans,'N',n,1,D,n,c,n,info)
    if (info/=0) then
       stop 'trisolve: Error in routine dtrtrs'
    end if
    trislv=pack(c,.TRUE.)
  end subroutine trisolve_v

  subroutine trisolve_m(A,b,trislv,trans)
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(:,:), intent(in):: B
    character, intent(in):: trans
    real(dp), dimension(size(A,2),size(B,2)), intent(out):: trislv
    real(dp), dimension(size(A,1),size(A,2)):: D
    real(dp), dimension(size(A,2),size(B,2)):: C
    integer:: n, info
    if(size(A,1)/=size(A,2)) then
       stop 'Matrix must be square'
    end if
    n=size(A,1)
    D=A
    C=B
    call dtrtrs('U',trans,'N',n,size(C,2),D,n,C,n,info)
    if (info/=0) then
       stop 'trisolve: Error in routine dtrtrs'
    end if
    trislv=C
  end subroutine trisolve_m

  subroutine symeigen(B,Eval,Evec)
    implicit none
    real(dp), intent(in):: B(:,:)
    real(dp), intent(out):: Eval(size(B,1)), Evec(size(B,1),size(B,1))
    real(dp):: A(size(B,1),size(B,1)), abstol, vl, vu
    integer:: lwork, liwork, na, info, il, iu, ne
    real(dp), allocatable:: work(:)
    integer, allocatable:: iwork(:), isuppz(:)
    A=B
    na=size(A,1)
    abstol=0.0_dp
    lwork=50*na
    liwork=10*na
    il=1
    iu=na
    vu=huge(1.0_dp)
    vl=-vu
    allocate(work(lwork),iwork(liwork),isuppz(2*na))
    call dsyevr('V','A','U',na,A,na,vl,vu,il,iu,abstol,ne,Eval,Evec,&
         & na,isuppz,work,lwork,iwork,liwork,info)
    deallocate(work,iwork,isuppz)
  end subroutine symeigen

  subroutine eigensymm(A,ev)
    ! Computes the eigen values and vectors for a symmetric real
    ! matrix
    implicit none
    real(dp), dimension(:,:), intent(inout):: A
    real(dp), dimension(:), intent(out):: ev
    call symeigen(A,ev,A)
  end subroutine eigensymm

  function qrR(A) result(R)
    ! Computes upper triangular matrix R of QR decomposition A=Q%*%R
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp):: R(size(A,2),size(A,2)), B(size(A,1),size(A,2)), tau(size(A,2))
    real(dp), allocatable:: work(:)
    integer:: m, n, lwork, info
    m=size(A,1)
    n=size(A,2)
    lwork=m*m
    allocate(work(lwork))
    B=A
    if (m<n) then
       stop 'qrR: # of rows must be > than the # of columns'
    end if
    call dgeqrf(m,n,B,m,tau,work,lwork,info)
    if (info/=0) then
       stop 'qrR: failure of routine dgeqrf'
    end if
    deallocate(work)
    R=B(1:n,1:n)
    do m=1,(n-1)
       R((m+1):n,m)=0.0_dp
    end do       
  end function qrR

  function ginv(A)
    ! Computes the M-P generalized inverse using SVD divide and conquer
    ! method
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,2),size(A,1)):: ginv
    real(dp), dimension(size(A,1),size(A,2)):: UU
    real(dp), dimension(size(A,2),size(A,2)):: VVT
    real(dp), dimension(size(A,2)):: s
    if(size(A,1)<size(A,2)) then
       stop 'No overdetermined systems to SVD'
    end if
    call svd(A,UU,s,VVT)
    where(abs(s) > 0.00000001_dp)
       s=1.0_dp/s
    elsewhere
       s=0.0_dp
    end where
    ginv=(VVT.tmmd.s).mmt.UU
  end function ginv

  subroutine svd(A,UU,SS,VVT)
    ! Computes the M-P generalized inverse using SVD divide and conquer
    ! method
    implicit none
    real(dp), dimension(:,:), intent(in):: A
    real(dp), dimension(size(A,1),size(A,2)), intent(out):: UU
    real(dp), dimension(size(A,2)), intent(out):: SS
    real(dp), dimension(size(A,2),size(A,2)), intent(out):: VVT
    real(dp), dimension(size(A,1),size(A,2)):: B
    real(dp), allocatable:: wrk(:)
    integer, allocatable:: iwrk(:)
    integer:: m, n, info, lwrk, mxdm, mndm
    m=size(A,1)
    n=size(A,2)
    mxdm=max(m,n)
    mndm=min(m,n)
    lwrk=6*mndm*mndm+max(mxdm,5*mndm*mndm+5*mndm)
    B=A
    allocate(wrk(lwrk),iwrk(8*mndm))
    call dgesdd('S',m,n,B,m,SS,UU,m,VVT,n,wrk,lwrk,iwrk,info)
    if (info/=0) then
       stop 'SVD: error in routine'
    end if
    deallocate(wrk,iwrk)
  end subroutine svd

  subroutine flsqr(X,y,coef,info)
    implicit none
    real(dp), dimension(:,:), intent(in):: X
    real(dp), dimension(:),intent(in):: y
    integer, intent(out):: info
    real(dp), dimension(size(X,2)), intent(out):: coef
    real(dp), allocatable:: work(:)
    real(dp), dimension(size(X,1),1):: ytmp
    real(dp), dimension(size(X,1),size(X,2)):: Xtmp
    integer:: m, n, lwrk, jpvt(size(X,2)), rank
    real(dp):: w(1)
    m=size(X,1)
    n=size(X,2)
    Xtmp=X
    ytmp=reshape(y,(/m,1/))
    lwrk=-1
    call dgelsy(m,n,1,Xtmp,m,ytmp,m,jpvt,1.0e-8_dp,rank,w,lwrk,info)
    lwrk=w(1)
    allocate(work(lwrk))
    call dgelsy(m,n,1,Xtmp,m,ytmp,m,jpvt,1.0e-8_dp,rank,work,lwrk,info)
    coef=ytmp(1:n,1)
    deallocate(work)
  end subroutine flsqr

  subroutine lsqr(X,y,coef,RSS)
    implicit none
    real(dp), dimension(:,:), intent(in):: X
    real(dp), dimension(:),intent(in):: y
    real(dp), dimension(size(X,2)), intent(out):: coef
    real(dp), intent(out), optional:: RSS
    real(dp), dimension(size(X,1),1):: ytmp
    real(dp), allocatable:: work(:)
    real(dp), dimension(size(X,1),size(X,2)):: Xtmp
    integer:: m, n, info, lwrk, jpvt(size(X,2)), rank
    real(dp):: w(1)
    m=size(X,1)
    n=size(X,2)
    if (m/=size(y)) then
       stop 'Rows of X and length of y must match'
    end if
    Xtmp=X
    ytmp=reshape(y,(/size(X,1),1/))
    lwrk=-1
    call dgelsy(m,n,1,Xtmp,m,ytmp,m,jpvt,1.0e-8_dp,rank,w,lwrk,info)
    lwrk=w(1)
    allocate(work(lwrk))
    call dgelsy(m,n,1,Xtmp,m,ytmp,m,jpvt,1.0e-8_dp,rank,work,lwrk,info)
    if(info/=0) then
       stop 'Problem with dgelsd L2 solution'
    end if
    coef=ytmp(1:n,1)
    if(present(RSS)) then
       RSS=sum((ytmp((size(X,2)+1):size(X,1),1))**2)
    end if
    deallocate(work)
  end subroutine lsqr

  function poisfitn(X,y)
    implicit none
    real(dp), dimension(:,:), intent(in):: X
    real(dp), dimension(:), intent(in):: y
    real(dp), dimension(size(X,1),size(X,2)):: W
    real(dp), dimension(size(y)):: mu, ywork,logy
    real(dp), dimension(size(X,2)):: poisfitn, parm, parmold
    real(dp):: checker
    integer:: i
    if (size(X,1)/=size(y)) then
       stop 'Rows of X and length of y must match'
    end if
    logy=y
    where (y<0.0001_dp)
       logy=0.01_dp
    end where
    logy=log(logy)
    call lsqr(X,logy,parm)
    do i=1,100
       parmold=parm
       mu=exp(X.mm.parm)
       ywork=(log(mu)+(y-mu)/mu)*sqrt(mu)
       W=X*spread(sqrt(mu),2,size(X,2))
       call lsqr(W,ywork,parm)
       checker=abs(sum(parm-parmold))
       if (checker <= 0.0000001_dp) then
          poisfitn=parm
          return
       end if 
    end do
    poisfitn=0.0_dp
  end function poisfitn

  function poisfito(X,y,offset)
    implicit none
    real(dp), dimension(:,:), intent(in):: X
    real(dp), dimension(:), intent(in):: y
    real(dp), dimension(:), intent(in):: offset
    real(dp), dimension(size(X,1),size(X,2)):: W
    real(dp), dimension(size(y)):: mu, ywork, logy
    real(dp), dimension(size(X,2)):: poisfito, parm, parmold
    real(dp):: checker
    integer:: i
    if (size(X,1)/=size(y)) then
       stop 'Rows of X and length of y must match'
    end if
    logy=y
    where (y<0.0001_dp)
       logy=0.01_dp
    end where
    if (size(y)/=size(offset)) then
       stop 'Offset must has wrong dimension'
    end if
    logy=log(logy) - offset
    call lsqr(X,logy,parm)
    do i=1,100
       parmold=parm
       mu=exp((X.mm.parm) + offset)
       ywork=(log(mu)-offset+(y-mu)/mu)*sqrt(mu)
       W=X*spread(sqrt(mu),2,size(X,2))
       call lsqr(W,ywork,parm)
       checker=abs(sum(parm-parmold))
       if (checker <= 0.0000001_dp) then
          poisfito=parm
          return
       end if
    end do
    poisfito=0.0_dp
  end function poisfito

end module matrix

module dm_util
  implicit none 

contains

subroutine newton(n,x,fvec,finfo,fvecjac,niter,ierr)
  use mytype
  use matrix, only: flsqr, fsymsolve, operator(.mm.)
  implicit none
  integer, intent(in):: n, niter
  real(dp), intent(inout):: x(n)
  real(dp), intent(out):: fvec(n), finfo(n,n)
  integer, intent(inout):: ierr
  integer:: i
  real(dp), parameter:: toler=1.0D-6
  real(dp):: s(n), xval, x_old(n)
  integer, parameter:: maxit=150
  interface
     subroutine fvecjac(n,x,fvec,fjac)
       use mytype
       implicit none
       integer, intent(in):: n
       real(dp), intent(in):: x(n)
       real(dp), intent(out):: fvec(n)
       real(dp), intent(out):: fjac(n,n)
     end subroutine fvecjac
  end interface
  ierr=0
  x_old=x
  do i=1,niter
     call fvecjac(n,x,fvec,finfo)
     if (ierr /= 0) return
     if (i >= 5 ) then
        call fsymsolve(finfo,fvec,s,ierr)
        if (ierr /= 0) then
           return
        end if
     else
        call flsqr(finfo,fvec,s,ierr)
        if (ierr /= 0) then
           return
        end if
     end if
     xval=maxval(abs(fvec))
     x=x+s
     if(xval<toler .or. maxval(abs(x-x_old))<toler) then
!        write(*,*) i
        return
     end if
     x_old=x
     if(i==maxit) then
        ierr=999
     end if
  end do
end subroutine newton

subroutine fmin(ax,bx,xmin,fval,fn,tol)
  use mytype
  implicit none
  real(dp), intent(in):: ax, bx, tol
  real(dp), intent(out):: fval, xmin
  interface
     function fn(x)
       use mytype
       implicit none
       real(dp), intent(in):: x
       real(dp):: fn
     end function fn
  end interface
!
!      an approximation  x  to the point where  f  attains a minimum  on
!  the interval  (ax,bx)  is determined.
!
!  input..
!
!  ax    left endpoint of initial interval
!  bx    right endpoint of initial interval
!  f     function subprogram which evaluates  f(x)  for any  x
!        in the interval  (ax,bx)
!  tol   desired length of the interval of uncertainty of the final
!        result (.ge.0.)
!
!  output..
!
!  fmin  abcissa approximating the point where  f  attains a
!        minimum
!
!      the method used is a combination of  golden  section  search  and
!  successive parabolic interpolation.  convergence is never much slower
!  than  that  for  a  fibonacci search.  if  f  has a continuous second
!  derivative which is positive at the minimum (which is not  at  ax  or
!  bx),  then  convergence  is  superlinear, and usually of the order of
!  about  1.324....
!      the function  f  is never evaluated at two points closer together
!  than  eps*abs(fmin)+(tol/3), where eps is  approximately  the  square
!  root  of  the  relative  machine  precision.   if   f   is a unimodal
!  function and the computed values of   f   are  always  unimodal  when
!  separated  by  at least  eps*abs(x)+(tol/3), then  fmin  approximates
!  the abcissa of the global minimum of  f  on the interval  ax,bx  with
!  an error less than  3*eps*abs(fmin)+tol.  if   f   is  not  unimodal,
!  then fmin may approximate a local, but perhaps non-global, minimum to
!  the same accuracy.
!      this function subprogram is a slightly modified  version  of  the
!  algol  60 procedure  localmin  given in richard brent, algorithms for
!  minimization without derivatives, prentice-hall, inc. (1973).
!
!
      real(dp):: a,b,c,d,e,eps,xm,p,q,r,tol1,t2,u,v,w,fu,fv,fw,fx,x,tol3
!
!  ! is the squared inverse of the golden ratio
      c=0.5_dp*(3.0d0-sqrt(5.0d0))
!
!  eps is approximately the square root of the relative machine
!  precision.
!
      eps=epsilon(1.0_dp)
      tol1=eps+1.0_dp
      eps=sqrt(eps)
!
      a=ax
      b=bx
      v=a+c*(b-a)
      w=v
      x=v
      e=0.0_dp
      fx=fn(x)
      fv=fx
      fw=fx
      tol3=tol/3.0d0
!
!  main loop starts here
!
20    xm=0.5_dp*(a+b)
      tol1=eps*abs(x)+tol3
      t2=2.0_dp*tol1
!
!  check stopping criterion
!
      if (abs(x-xm).le.(t2-0.5_dp*(b-a))) go to 190
      p=0.0_dp
      q=0.0_dp
      r=0.0_dp
      if (abs(e).le.tol1) go to 50
!
!  fit parabola
!
      r=(x-w)*(fx-fv)
      q=(x-v)*(fx-fw)
      p=(x-v)*q-(x-w)*r
      q=2.0_dp*(q-r)
      if (q.le.0.0_dp) go to 30
      p=-p
      go to 40
30    q=-q
40    r=e
      e=d
50    if ((abs(p).ge.abs(0.5_dp*q*r)).or.(p.le.q*(a-x)).or.(p.ge.q*(b-x))) go to 60
!
!  a parabolic-interpolation step
!
      d=p/q
      u=x+d
!
!  f must not be evaluated too close to ax or bx
!
      if (((u-a).ge.t2).and.((b-u).ge.t2)) go to 90
      d=tol1
      if (x.ge.xm) d=-d
      go to 90
!
!  a golden-section step
!
60    if (x.ge.xm) go to 70
      e=b-x
      go to 80
70    e=a-x
80    d=c*e
!
!  f must not be evaluated too close to x
!
90    if (abs(d).lt.tol1) go to 100
      u=x+d
      go to 120
100   if (d.le.0.0_dp) go to 110
      u=x+tol1
      go to 120
110   u=x-tol1
120   fu=fn(u)
!
!  update  a, b, v, w, and x
!
      if (fx.gt.fu) go to 140
      if (u.ge.x) go to 130
      a=u
      go to 140
130   b=u
140   if (fu.gt.fx) go to 170
      if (u.ge.x) go to 150
      b=x
      go to 160
150   a=x
160   v=w
      fv=fw
      w=x
      fw=fx
      x=u
      fx=fu
      go to 20
170   if ((fu.gt.fw).and.(w.ne.x)) go to 180
      v=w
      fv=fw
      w=u
      fw=fu
      go to 20
180   if ((fu.gt.fv).and.(v.ne.x).and.(v.ne.w)) go to 20
      v=u
      fv=fu
      go to 20
!
!  end of main loop
!
190   fval=fn(x)  
      xmin=x
      return
end subroutine fmin

subroutine set_random_seed
! Use time stamp to init the random number generator.  For many f95 compilers this is done
! using the intrisic subroutine random_seed() but this behavior isn't enforced by the 
! standard so this function give a "universal" method of seeding the RNG that should work on
! any f95 standard complying compiler.
  implicit none
  integer :: idate(8)
  integer, allocatable :: iseed(:)
  call date_and_time(VALUES=idate)
  allocate( iseed(8) )
  iseed = idate(8)
  call random_seed(PUT=iseed)
  deallocate ( iseed )  
end subroutine set_random_seed

subroutine diagadd(MM,dd)
  use mytype
  implicit none 
  real(dp), intent(inout):: MM(:,:)
  real(dp), intent(in):: dd(:)
  integer:: j,n
  n = size(MM,1)
  if ((size(MM,2) /= n) .or. (size(dd) /= n)) then
     stop 'diagadd: dimension of inputs do not match!'
  end if
  do j=1,n
     MM(j,j)=MM(j,j)+dd(j)
  end do
end subroutine diagadd

function outerprod(a,b) result(oprod)
  use mytype
  implicit none
  real(dp), intent(in):: a(:),b(:)
  real(dp), dimension(size(a),size(b)) :: oprod
  integer:: i
  do i=1,size(b)
     oprod(:,i)=a*b(i)
  end do
end function outerprod

function bbase(x,xl,xu,nknt,deg) result(Bm)
  use mytype
  use matrix, only: operator(.mmt.)
  implicit none
  real(dp), intent(in):: x(:), xl, xu
  integer, intent(in):: nknt, deg
  real(dp):: Bm(size(x),nknt+deg+1), kseq(nknt+2*deg+2), &
       & Dm(nknt+deg+1,nknt+2*deg+2), tBm(size(x),nknt+2*deg+2)
  integer:: i, tdm, fact
  fact=1.0_dp
  do i=1,deg
     fact=fact*i
  end do
  tdm=nknt+2*deg+2
  kseq=(/(real(i,dp),i=1,tdm)/)
  kseq=kseq-kseq(deg+1)
  kseq=kseq/kseq(nknt+deg+2)
  kseq=(xu-xl)*kseq+xl
  Dm=diffd(tdm,deg+1)
  do i=1,tdm
     tBm(:,i)=x-kseq(i)
  end do
  where (tBm <= 0.0_dp)
     tBm=0.0_dp
  elsewhere
     tBm=tBm**deg
  end where
  Bm=tBm.mmt.Dm
  Bm=(((-1.0_dp)**(deg+1))/(real(fact,dp)*(kseq(2)-kseq(1))**deg))*Bm
end function bbase

function diffd(dim,dif) result(Dm)
  use mytype
  use matrix, only: eye
  implicit none
  integer, intent(in):: dim, dif
  real(dp):: Dm(dim-dif,dim)
  real(dp):: Em(dim,dim)
  integer:: i
  Em=eye(dim)
  do i=1,dif
     Em(1:(dim-i),:)=Em(2:dim-i+1,:)-Em(1:(dim-i),:)
  end do
  Dm=Em(1:(dim-dif),:)
end function diffd

function poisson (x,expo)
  use mytype
  implicit none
  real(dp), intent(in):: x,expo
  real(dp):: poisson
  poisson=-expo+x*log(expo)
  return
end function poisson

function lgamma(x,kf) result(gl)
!
!       ==================================================
!       Purpose: Compute gamma (x) or (x)]
!       Input:   x  --- Argument (x) ( x > 0 )
!                KF --- Function code
!                       KF=1 a(x); KF=0 for ln[a(x)]
!       Output:  GL --- a(x) or ln[a(x)] 
!       ==================================================
!
  use mytype
  implicit none
  real(dp), intent(in):: x
  integer, intent(in):: kf
  real(dp):: gl
  real(dp), dimension(10),parameter:: a=(/8.333333333333333e-02_dp, &
       & -2.777777777777778e-03_dp, 7.936507936507937e-04_dp, &
       & -5.952380952380952e-04_dp, 8.417508417508418e-04_dp, &
       & -1.917526917526918e-03_dp, 6.410256410256410e-03_dp, &
       & -2.955065359477124e-02_dp, 1.796443723688307e-01_dp, &
       & -1.39243221690590_dp/)
  real(dp):: x0, x2, xp, glo
  integer:: k, n
  x0=x
  if (x == 1.0 .OR. x == 2.0) then
     gl=0.0_dp
     if (kf == 1) gl=exp(gl)
     return
  else if (x <= 7.0) then
     n=int(7-x)
     x0=x+n
  end if
  x2=1.0_dp/(x0*x0)
  xp=6.283185307179586477_dp
  glo=A(10)
  do k=9,1,-1
     glo=glo*x2+a(k)
  end do
  gl=glo/x0+0.5_dp*log(xp)+(x0-.5_dp)*log(x0)-x0
  if (x <= 7.0) then
     do k=1,n
        gl=gl-log(x0-1.0_dp)
        x0=x0-1.0_dp
     end do
  end if
  if (kf == 1) gl=exp(gl)
  return
end function lgamma

function gammln(x) result(gmln)
  use mytype
  implicit none
  real(dp), intent(in):: x(:)
  real(dp):: gmln(size(x))
  integer:: nx, i
  nx=size(x)
  do i=1,nx
     gmln(i)=lgamma(x(i),0)
  end do
end function gammln

end module dm_util
